/**
 * 
 */
package prompt.multiperspective.externals.conformance;

import info.clearthought.layout.TableLayout;
import info.clearthought.layout.TableLayoutConstants;

import java.awt.Dimension;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.swing.JComboBox;
import javax.swing.JPanel;

import org.deckfour.xes.classification.XEventClass;
import org.deckfour.xes.classification.XEventClasses;
import org.deckfour.xes.classification.XEventClassifier;
import org.deckfour.xes.info.XLogInfo;
import org.deckfour.xes.info.XLogInfoFactory;
import org.deckfour.xes.model.XLog;
import org.processmining.framework.util.ArrayUtils;
import org.processmining.framework.util.Pair;
import org.processmining.models.graphbased.directed.petrinet.PetrinetGraph;
import org.processmining.models.graphbased.directed.petrinet.elements.Transition;
import org.processmining.plugins.connectionfactories.logpetrinet.TransEvClassMapping;

import uk.ac.shef.wit.simmetrics.similaritymetrics.AbstractStringMetric;
import uk.ac.shef.wit.simmetrics.similaritymetrics.Levenshtein;

import com.fluxicon.slickerbox.factory.SlickerFactory;

/**
 * GUI to map event class (with any classifiers) to transitions of Petri net
 * 
 * @author aadrians
 * 
 */
public class PromptEvClassLogPetrinetConnectionFactoryUI extends JPanel {
	private static final long serialVersionUID = -699953189980632566L;
	
	private int rowCounter;

	// dummy event class (for unmapped transitions)
	public final static XEventClass DUMMY = new XEventClass("DUMMY", -1) {
		public boolean equals(Object o) {
			return this == o;
		}
	};

	// internal attributes
	@SuppressWarnings("unused")
	private final XLog log;
	private Map<Transition, JComboBox> mapTrans2ComboBox = new HashMap<Transition, JComboBox>();
	private JComboBox classifierSelectionCbBox;
	private Object choosenClassifier = null;
	private PetrinetGraph net = null;

	public PromptEvClassLogPetrinetConnectionFactoryUI(final XLog log, PetrinetGraph net, Object[] availableClassifier) {
		super();

		// index for row
		rowCounter = 0;

		// import variable
		this.log = log;
		this.net = net;

		// swing factory
		SlickerFactory factory = SlickerFactory.instance();

		// set layout
		double size[][] = { { TableLayoutConstants.FILL, TableLayoutConstants.FILL }, { 80, 70 } };
		TableLayout layout = new TableLayout(size);
		setLayout(layout);

		// label
		add(factory
				.createLabel("<html><h1>Map Transitions to Event Class</h1><p>First, select a classifier. Unmapped transitions will be mapped to a dummy event class.</p></html>"),
				"0, " + rowCounter + ", 1, " + rowCounter);
		rowCounter++;

		// add classifier selection
		classifierSelectionCbBox = factory.createComboBox(availableClassifier);
		

		final String invisibleTransitionRegEx = "[a-z][0-9]+|(tr[0-9]+)|(silent)|(tau)|(skip)|(invi)";
		final Pattern pattern = Pattern.compile(invisibleTransitionRegEx);
		
		Pair<Object, Float> ccls = this.getBestClassifier(availableClassifier, pattern);
		choosenClassifier = ccls.getFirst();
		float maxAverage = ccls.getSecond();

		classifierSelectionCbBox.setSelectedItem(factory.createComboBox(availableClassifier));

		// add mapping between transitions and selected event class 
		Object[] boxOptions = extractEventClasses(log, (XEventClassifier) choosenClassifier);
		List<Transition> listTrans = new ArrayList<Transition>(net.getTransitions());
		Collections.sort(listTrans, new Comparator<Transition>() {
			public int compare(Transition o1, Transition o2) {
				return o1.getLabel().compareTo(o2.getLabel());
			}
		});
		for (Transition transition : listTrans) {

			Pair<Integer, Float> selected =
					preSelectOption(
						transition.getLabel().toLowerCase(),
						boxOptions,
						pattern);
			
			JComboBox cbBox = factory.createComboBox(boxOptions);
			cbBox.setPreferredSize(new Dimension(350, 30));
			cbBox.setMinimumSize(new Dimension(350, 30));
			mapTrans2ComboBox.put(transition, cbBox);
			
			if (transition.isInvisible()) {
				cbBox.setSelectedItem(DUMMY);
			} else {
				cbBox.setSelectedIndex(selected.getFirst());
			}
			
			// we show the mapping dropdown box only if the transition is not
			// invisible or if its distance from the activities to whom it has
			// been mapped is below the average
			if (!transition.isInvisible() && (selected.getSecond()) < maxAverage) {
				layout.insertRow(rowCounter, 30);
				
				add(factory.createLabel(transition.getLabel()), "0, "
						+ rowCounter + ", l, c");
				add(cbBox, "1, " + rowCounter + ", l, c");
				rowCounter++;
			}
		}
	}
	/**
	 * Returns the best classifier in terms of overall similarity between
	 * members of the matching between transitions and events
	 * 
	 * @param net	the Petri network to be analyzed
	 * @param availableClassifier	array of all available classifiers
	 * @param pattern	pattern to match invisible transitions
	 * @return 	a pair containing the best classifier and the average similarity
	 * 			scored
	 */
	public Pair<Object, Float> getBestClassifier(Object[] availableClassifier, Pattern pattern){
		// store the choosen classifier locally
		Object cls = null;
		// store the value of the maximum similarity measured for a classifier
		float maxAverage = 0;
		// calculate average similarity of transition and activity names
		// for each classifier	
		for (Object o : availableClassifier){
			// array of all events strings using the specified classifier
			Object[] boxOptions = extractEventClasses(log, (XEventClassifier) o);
			// list of all transitions in the network
			List<Transition> listTrans = new ArrayList<Transition>(this.net.getTransitions());
			
			float total = 0;
			float average = 0;
			int counter = 0;
			// calculate average similarity of transition and activity names
			// with the present classifier
			for (Transition transition : listTrans) {
				
				Pair<Integer, Float> selected =
						preSelectOption(
							transition.getLabel().toLowerCase(),
							boxOptions,
							pattern);
				// invisible transition are associated with dummy activity by
				// default
				if (!transition.isInvisible()) {
					total = total + selected.getSecond();
					counter++;
				}
			}
			if (counter > 0)
				average = total/counter;
			else
				average = 0;
			
			if(average > maxAverage){
				maxAverage = average;
				cls = o;
			}
		}
		
		return new Pair<Object, Float>(cls, maxAverage);
	}

	/**
	 * get all available event classes using the selected classifier, add with
	 * NONE
	 * 
	 * @param log
	 * @param selectedItem
	 * @return
	 */
	private Object[] extractEventClasses(XLog log, XEventClassifier selectedItem) {
		XLogInfo summary = XLogInfoFactory.createLogInfo(log, selectedItem);
		XEventClasses eventClasses = summary.getEventClasses();

		// sort event class
		Collection<XEventClass> classes = eventClasses.getClasses();

		// create possible event classes
		Object[] arrEvClass = classes.toArray();
		Arrays.sort(arrEvClass);
		Object[] notMappedAct = { "NONE" };
		Object[] boxOptions = ArrayUtils.concatAll(notMappedAct, arrEvClass);

		return boxOptions;
	}

	/**
	 * Returns the Event Option Box index of the most similar event for the
	 * transition.
	 * 
	 * @param transition
	 *            Name of the transitions, assuming low cases
	 * @param events
	 *            Array with the options for this transition
	 * @return Index of option more similar to the transition
	 */
	private Pair<Integer, Float> preSelectOption(
		String transition,
		Object[] events,
		Pattern pattern) {
		// try to find precise match
		for (int i = 1; i < events.length; i++) {
			String event = ((XEventClass) events[i]).toString().toLowerCase();
			if (event.equalsIgnoreCase(transition)) {
				return new Pair<Integer, Float>(i, Float.parseFloat("1"));
			}
			;
		}

		Matcher matcher = pattern.matcher(transition);
		if (matcher.find() && matcher.start() == 0) {
			return new Pair<Integer, Float>(0, new Float(0.0));
		}

		// The metric to get the similarity between strings
		AbstractStringMetric metric = new Levenshtein();

		int index = 0;
		float simOld = Float.MIN_VALUE;
		for (int i = 1; i < events.length; i++) {
			String event =
					((XEventClass) events[i]).toString().toLowerCase();
			float sim = metric.getSimilarity(transition, event);

			if (simOld < sim) {
				simOld = sim;
				index = i;
			}
		}

		return new Pair<Integer, Float>(index, simOld);
	}

	/**
	 * Generate the map between Transitions and Event according to the user
	 * selection.
	 * 
	 * @return Map between Transitions and Events.
	 */
	public TransEvClassMapping getMap() {
		TransEvClassMapping map = new TransEvClassMapping(
				(XEventClassifier) choosenClassifier, DUMMY);
		for (Transition trans : mapTrans2ComboBox.keySet()) {
			Object selectedValue = mapTrans2ComboBox.get(trans).getSelectedItem();
			if (selectedValue instanceof XEventClass) {
				// a real event class
				map.put(trans, (XEventClass) selectedValue);
			} else {
				// this is "NONE"
				map.put(trans, DUMMY);
			}
		}
		return map;
	}

	/**
	 * Get the selected classifier (choosen automatically or selected by the user)
	 * 
	 * @return selected classifier
	 */
	public XEventClassifier getSelectedClassifier() {
		return (XEventClassifier) choosenClassifier;
	}


	public int getRowCounter()
	{
		return rowCounter;
	}
	
	public void setRowCounter(int rowCounter)
	{
		this.rowCounter = rowCounter;
	}
}
