package prompt.multiperspective.plugins;

import java.util.Set;

import javax.swing.JPanel;

import org.deckfour.uitopia.api.event.TaskListener.InteractionResult;
import org.deckfour.xes.model.XLog;
import org.processmining.acceptingpetrinet.connections.MergeAcceptingPetriNetArrayIntoAcceptingPetriNetConnection;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNet;
import org.processmining.acceptingpetrinet.models.AcceptingPetriNetArray;
import org.processmining.acceptingpetrinetminer.connections.DiscoverAcceptingPetriNetArrayFromEventLogArrayConnection;
import org.processmining.contexts.uitopia.UIPluginContext;
import org.processmining.contexts.uitopia.annotations.UITopiaVariant;
import org.processmining.datadiscovery.model.DiscoveredDataPetriNet;
import org.processmining.datapetrinets.DataPetriNet;
import org.processmining.framework.connections.ConnectionCannotBeObtained;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.framework.providedobjects.ProvidedObjectManager;
import org.processmining.loclog.utils.BPMNUtils;
import org.processmining.log.models.EventLogArray;
import org.processmining.logalignment.models.LogAlignment;
import org.processmining.logalignment.models.LogAlignmentArray;
import org.processmining.logalignment.models.ReplayResultArray;
import org.processmining.models.connections.petrinets.behavioral.FinalMarkingConnection;
import org.processmining.models.connections.petrinets.behavioral.InitialMarkingConnection;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.petrinet.Petrinet;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithDataFactory;
import org.processmining.models.heuristics.HeuristicsNet;
import org.processmining.models.semantics.petrinet.Marking;
import org.processmining.plugins.InductiveMiner.mining.MiningParametersIMi;
import org.processmining.plugins.InductiveMiner.plugins.IMPetriNet;
import org.processmining.plugins.converters.BPMNConversionConnection;
import org.processmining.plugins.petrinet.replayresult.PNRepResult;
import org.processmining.plugins.replayer.replayresult.SyncReplayResult;
import org.processmining.processtree.ProcessTree;

import prompt.multiperspective.configuration.MultiperspectiveConfiguration;
import prompt.multiperspective.configuration.MultiperspectiveConfiguration.ACTIVITY;
import prompt.multiperspective.configuration.MultiperspectiveConfiguration.CONTROL_FLOW_MINER;
import prompt.multiperspective.configuration.MultiperspectiveConfiguration.MINER_CONFIGURATION;
import prompt.multiperspective.configuration.gui.Introduction;
import prompt.multiperspective.configuration.gui.SelectMiner;
import prompt.multiperspective.configuration.gui.SelectPerspectives;
import prompt.multiperspective.configuration.gui.SimpleConfiguration;
import prompt.multiperspective.externals.decisionminer.CustomDecisionMiner;
import prompt.multiperspective.utils.AuthorsInfo;
import prompt.multiperspective.utils.ContextUtils;
import prompt.multiperspective.utils.PluginNames;
import prompt.roles.data.RoleSet;

 
/**
 * This class contains the actual multiperspective miner
 * 
 * @author Andrea Burattin
 * @author Anna Kalenkova
 */
public class MultiperspectiveMiner {

	private static String CONFIGURATION_TITLE = "Multiperspective Miner";
	
	@Plugin(
		name = "Multiperspective Miner",
		parameterLabels = { "Event Log" },
		returnLabels = { "Multiperspective Model" },
		returnTypes = { MultiperspectiveModel.class },
		userAccessible = true
	)
	@UITopiaVariant(
		author = AuthorsInfo.AUTHOR,
		email = AuthorsInfo.EMAIL,
		affiliation = AuthorsInfo.AFFILIATION
	)
	public static MultiperspectiveModel mainPlugin(UIPluginContext context, XLog log) throws ConnectionCannotBeObtained {
		
		MultiperspectiveConfiguration configuration = new MultiperspectiveConfiguration();
		JPanel[] steps = {
			new Introduction(),
			new SelectPerspectives(configuration),
			new SelectMiner(configuration)
		};
		
		InteractionResult result = InteractionResult.NEXT;
		int currentStep = 0;
		int nofSteps = steps.length;
		boolean configurationOngoing = true;
		while (configurationOngoing && currentStep < nofSteps) {
			result = context.showWizard(CONFIGURATION_TITLE, currentStep == 0, false, steps[currentStep]);
			switch (result) {
				case NEXT :
					currentStep++;
					break;
				case PREV :
					currentStep--;
					break;
				case FINISHED :
					configurationOngoing = false;
					break;
				case CANCEL :
					return null;
				default :
					configurationOngoing = false;
					break;
			}
			if(currentStep == 2) {
				configuration = ((SelectPerspectives) steps[1]).getConfiguration();
				// Only if the advanced configuration was chosen, we proceed
				if(configuration.getMinerConfiguration() != MINER_CONFIGURATION.ADVANCED) {
					break;
				}
			}
		}
		if (result != InteractionResult.NEXT) {
			return null;
		}
		
		BPMNDiagram bpmnModel = null;
		Petrinet petriNet = null;
		RoleSet roles = null;
		DataPetriNet dataPetriNet = null;
		DataPetriNet dataPetriNetNoVars = null;
		//AdvancedSocialNetwork adSNA = null;
		
		long miningTime = 0l;
		long mergeTime = 0l;
		long miningSpace = 0l;
		long mergeSpace = 0l;
		
		// setting context
		context.getProgress().setMinimum(0);
		context.getProgress().setMaximum(configuration.getActivities().size());
		if (configuration.getActivities().contains(ACTIVITY.ROLE_DISCOVERY)) {
			context.getProgress().setMaximum(context.getProgress().getMaximum() + 1);
		}
		
		if (configuration.getMinerConfiguration() == MINER_CONFIGURATION.SIMPLE) {
			
			// SIMPLE CONFIGURATION
			// ====================
			
			SimpleConfiguration simpleConfiguration = new SimpleConfiguration();
			result = context.showConfiguration(CONFIGURATION_TITLE, simpleConfiguration);
			if (result != InteractionResult.CONTINUE) {
				return null;
			}
			
			// Inductive Miner
			long timer;
			long spacer;
			if (configuration.getActivities().contains(ACTIVITY.CONTROL_FLOW_DISCOVERY)) {
				timer = System.currentTimeMillis();
				spacer = usedSpace();
				context.log("Mining control-flow...");
				float fitness = simpleConfiguration.getFitness();
				MiningParametersIMi parameters = new MiningParametersIMi();
				parameters.setNoiseThreshold(fitness);
				Object[] results = IMPetriNet.minePetriNet(context, log, parameters);
				petriNet = (Petrinet) results[0];
				
				PetriNetWithDataFactory factory1 = new PetriNetWithDataFactory(petriNet, petriNet.getLabel());
				Marking[] markingsNoVars1 = factory1.cloneInitialAndFinalConnection(context);
				ContextUtils.addToContext(context, markingsNoVars1[0], "Initial marking no vars");
				ContextUtils.addToContext(context, markingsNoVars1[1], "Final marking no vars");
				dataPetriNet = factory1.getRetValue();
				
				PetriNetWithDataFactory factory2 = new PetriNetWithDataFactory(petriNet, petriNet.getLabel());
				Marking[] markingsNoVars2 = factory2.cloneInitialAndFinalConnection(context);
				ContextUtils.addToContext(context, markingsNoVars2[0], "Initial marking no vars");
				ContextUtils.addToContext(context, markingsNoVars2[1], "Final marking no vars");
				dataPetriNetNoVars = factory2.getRetValue();
				
				ContextUtils.addToContext(context, petriNet, "Mined Petri Net");
				ContextUtils.addToContext(context, (Marking) results[1], "Initial marking");
				ContextUtils.addToContext(context, (Marking) results[2], "Final marking");
				
				context.addConnection(new InitialMarkingConnection(petriNet, (Marking) results[1]));
				context.addConnection(new FinalMarkingConnection(petriNet, (Marking) results[2]));
				
				bpmnModel = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class, PluginNames.PETRI_NET_TO_BPMN, null, null, petriNet);
				miningTime += (System.currentTimeMillis() - timer);
				miningSpace += usedSpace() - spacer;
				System.out.println("Time control-flow discovery: " + (System.currentTimeMillis() - timer));
			}
			
			// Data Discovery
			if (configuration.getActivities().contains(ACTIVITY.GUARDS_DISCOVERY)) {
				timer = System.currentTimeMillis();
				spacer = usedSpace();
				context.log("Mining data perspective...");
				Object[] results = null;
				try {
					results = CustomDecisionMiner.plugin(context, petriNet, log, false);
				} catch (Exception e) {
					e.printStackTrace();
				}
				miningTime += (System.currentTimeMillis() - timer);
				miningSpace += usedSpace() - spacer;
				System.out.println("Time data discovery: " + (System.currentTimeMillis() - timer));
				
				if (result != null) {
					dataPetriNet = (PetriNetWithData) results[0];
					
					ProvidedObjectManager pom = context.getProvidedObjectManager();
					pom.createProvidedObject("Petri Net With Data", dataPetriNet, DataPetriNet.class, context);
					context.getGlobalContext().getResourceManager().getResourceForInstance(dataPetriNet);
					
					pom.createProvidedObject("marking1", (Marking) results[1], Marking.class, context);
					context.getGlobalContext().getResourceManager().getResourceForInstance((Marking) results[1]);
					pom.createProvidedObject("marking2", (Marking) results[2], Marking.class, context);
					context.getGlobalContext().getResourceManager().getResourceForInstance((Marking) results[2]);
					
					timer = System.currentTimeMillis();
					spacer = usedSpace();
					BPMNDiagram bpmnModelWithData = context.tryToFindOrConstructFirstObject(BPMNDiagram.class, null, null, dataPetriNet);
					if (bpmnModelWithData != null) {
						bpmnModel = bpmnModelWithData;
					}
					mergeTime += (System.currentTimeMillis() - timer);
					mergeSpace += usedSpace() - spacer;
					
					// prepare the data petrinet with no variables
					PetriNetWithDataFactory factory = new PetriNetWithDataFactory(dataPetriNet, dataPetriNet.getLabel());
					Marking[] markingsNoVars = factory.cloneInitialAndFinalConnection(context);
					ContextUtils.addToContext(context, markingsNoVars[0], "Initial marking no vars");
					ContextUtils.addToContext(context, markingsNoVars[1], "Final marking no vars");
					dataPetriNetNoVars = factory.getRetValue();
					ContextUtils.addToContext(context, dataPetriNetNoVars, "Petri Net With Data (without variables)");
				}
				context.getProgress().inc();
			}
			
			// Social perspective (roles)
			if (configuration.getActivities().contains(ACTIVITY.ROLE_DISCOVERY)) {
				timer = System.currentTimeMillis();
				spacer = usedSpace();
				context.log("Mining roles...");
				roles = context.tryToFindOrConstructFirstNamedObject(RoleSet.class, PluginNames.SWIMLANE_DISCOVERY, null, null, bpmnModel, log);
				miningTime += (System.currentTimeMillis() - timer);
				miningSpace += usedSpace() - spacer;
				System.out.println("Time roles discovery: " + (System.currentTimeMillis() - timer));
				
				timer = System.currentTimeMillis();
				spacer = usedSpace();
				if (roles != null) {
					BPMNDiagram bpmnModelWithSwimlanes = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class, PluginNames.DECORATE_BPMN_WITH_SWIMLANES, null, null, bpmnModel, roles);
					if (bpmnModelWithSwimlanes != null) {
						bpmnModel = bpmnModelWithSwimlanes;
					}
				}
				context.getProgress().inc();
				mergeTime += (System.currentTimeMillis() - timer);
				mergeSpace += usedSpace() - spacer;
			}
			
			// Social perspective (social network analysis)
			/*if (configuration.getActivities().contains(ACTIVITY.ADVANCED_SNA)) {
				context.log("Mining Advanced Social Network...");
				
				try {
					adSNA = Orchestrator.mineLight(context, log);
				} catch (IOException e) {
					e.printStackTrace();
				}
				
				context.getProgress().inc();
			}*/
			
		} else {
			
			// ADVANCED OR LOCALIZED LOGS CONFIGURATION
			// ======================
			try {
				// Replay result for data discovery
				PNRepResult replayResult = null;
				// Control-flow Miner
				if (configuration.getActivities().contains(ACTIVITY.CONTROL_FLOW_DISCOVERY)) {
					context.log("Mining control-flow...");
					if(configuration.getMinerConfiguration() == MINER_CONFIGURATION.ADVANCED) {
						// ADVANCED CONFIGURATION
						// ======================
						CONTROL_FLOW_MINER controlFlowMiner = configuration.getControlFlowMiner();

						switch (controlFlowMiner) {
							case INDUCTIVE_MINER :
								petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,
										PluginNames.PETRI_NET_INDUCTIVE_DISCOVERY, null, null, log);
								break;
							case ALPHA_MINER :
								petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,
										PluginNames.PETRI_NET_ALPHA_DISCOVERY, null, null, log);
								break;
							case HEURISTIC_MINER :
								HeuristicsNet hn = context.tryToFindOrConstructFirstNamedObject(HeuristicsNet.class,
										PluginNames.HEURISTIC_NET_DISCOVERY, null, null, log);
								petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,
										PluginNames.HEURISTIC_NET_TO_PETRI_NET, null, null, hn);
								break;
							case ILP_MINER :
								petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,
										PluginNames.PETRI_NET_ILP_DISCOVERY, null, null, log);
								break;
							case EVOLUTIONARY_TREE_MINER :
								ProcessTree pt = context.tryToFindOrConstructFirstNamedObject(ProcessTree.class,
										PluginNames.PROCESS_TREE_ETM_DISCOVERY, null, null, log);
								petriNet = context.tryToFindOrConstructFirstNamedObject(Petrinet.class,
										PluginNames.PROCESS_TREE_TO_PETRI_NET, null, null, pt);
								break;
						}
						bpmnModel = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class,
								PluginNames.PETRI_NET_TO_BPMN, null, null, petriNet);
					} else {
						// LOCALIZED LOGS CONFIGURATION
						// @author Anna Kalenkova
						// ======================
						// Retrieving a Petri net
						context.log("Mining subprocesses...");
						AcceptingPetriNet apn;
						try {
							apn = context.tryToFindOrConstructFirstNamedObject(AcceptingPetriNet.class,
									PluginNames.MINE_PETRI_NET_USING_LOCALIZED_INFO, null, null, log);
							
							petriNet = apn.getNet();
							Marking initialMarking = apn.getInitialMarking();
							Set<Marking> finalMarkings = apn.getFinalMarkings();
							
							context.getProvidedObjectManager().createProvidedObject("Initial marking for " + petriNet.getLabel(),
									apn.getInitialMarking(), Marking.class, context);
							context.addConnection(new InitialMarkingConnection(petriNet, initialMarking));

							for (Marking finalMarking : finalMarkings) {
								context.getProvidedObjectManager().createProvidedObject("Final marking for " + petriNet.getLabel(),
										finalMarking, Marking.class, context);
								context.addConnection(new FinalMarkingConnection(petriNet, finalMarking));
							}
						
							Thread.sleep(10000);
							bpmnModel = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class,
									PluginNames.CONSTRUCT_BPMN_FROM_LOCALIZED_INFO, BPMNConversionConnection.class, 
									BPMNConversionConnection.BPMN_DIAGRAM, petriNet, log);

							// Calculate an alignment using divide and conquer
							if (configuration.getActivities().contains(ACTIVITY.GUARDS_DISCOVERY)) {
								MergeAcceptingPetriNetArrayIntoAcceptingPetriNetConnection mergeConnection = context
										.getConnectionManager().getFirstConnection(
												MergeAcceptingPetriNetArrayIntoAcceptingPetriNetConnection.class, context,
												apn);
								AcceptingPetriNetArray petriNetArray = mergeConnection
										.getObjectWithRole(MergeAcceptingPetriNetArrayIntoAcceptingPetriNetConnection.NETS);
								DiscoverAcceptingPetriNetArrayFromEventLogArrayConnection eventLogsConnection = context
										.getConnectionManager().getFirstConnection(
												DiscoverAcceptingPetriNetArrayFromEventLogArrayConnection.class, context,
												petriNetArray);
								EventLogArray eventLogArray = eventLogsConnection
										.getObjectWithRole(DiscoverAcceptingPetriNetArrayFromEventLogArrayConnection.LOGS);
								ReplayResultArray replayArray = context.tryToFindOrConstructFirstNamedObject(
										ReplayResultArray.class, PluginNames.REPLAY_EVENT_LOGS, null, null, petriNetArray,
										eventLogArray);
								LogAlignmentArray alignmentArray = context.tryToFindOrConstructFirstNamedObject(
										LogAlignmentArray.class, PluginNames.CONVERT_TO_LOG_ALIGNMENTS, null, null,
										replayArray, petriNetArray, eventLogArray);
								LogAlignment alignment = context.tryToFindOrConstructFirstNamedObject(LogAlignment.class,
										PluginNames.MERGE_LOG_ALIGNMENTS, null, null, alignmentArray, log, petriNetArray);
								replayResult = context.tryToFindOrConstructFirstNamedObject(PNRepResult.class,
										PluginNames.CONVERT_TO_REPLAY_RESULTS, null, null, alignment, apn, log);
								replayResult.getInfo().put(PNRepResult.TRACEFITNESS, new Double(1.0));
								for (SyncReplayResult traceReplay : replayResult) {
									traceReplay.getInfo().put(PNRepResult.TRACEFITNESS, new Double(1.0));
								}
							}
						} catch (Exception e) {
							e.printStackTrace();
						}
					}

					// Data Discovery
					if (configuration.getActivities().contains(ACTIVITY.GUARDS_DISCOVERY)) {
						dataPetriNet = new PetriNetWithDataFactory(petriNet, petriNet.getLabel()).getRetValue();
						dataPetriNetNoVars = new PetriNetWithDataFactory(petriNet, petriNet.getLabel()).getRetValue();
						context.getProgress().inc();
						context.log("Mining data perspective...");
						System.out.println("---------- " + replayResult);
						dataPetriNet = context.tryToFindOrConstructFirstNamedObject(DiscoveredDataPetriNet.class,
									PluginNames.DISCOVER_DATA_FLOW, null, null, petriNet, log, replayResult);
						ProvidedObjectManager pom = context.getProvidedObjectManager();
						pom.createProvidedObject("pnwd", dataPetriNet, DataPetriNet.class, context);
						context.getGlobalContext().getResourceManager().getResourceForInstance(dataPetriNet);

						BPMNDiagram bpmnModelWithData = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class,
								PluginNames.DATA_PETRI_NET_TO_BPMN, BPMNConversionConnection.class,
								BPMNConversionConnection.BPMN_DIAGRAM, dataPetriNet);
						if (bpmnModelWithData != null) {
							bpmnModel = bpmnModelWithData;
						}
						BPMNUtils.replaceActivitiesByEvents(bpmnModel);

						// prepare the data petrinet with no variables
						PetriNetWithDataFactory factory = new PetriNetWithDataFactory(dataPetriNet,
								dataPetriNet.getLabel());
						Marking[] markingsNoVars = factory.cloneInitialAndFinalConnection(context);
						ContextUtils.addToContext(context, markingsNoVars[0], "Initial marking no vars");
						ContextUtils.addToContext(context, markingsNoVars[1], "Final marking no vars");
						dataPetriNetNoVars = factory.getRetValue();
						ContextUtils.addToContext(context, dataPetriNetNoVars,
								"Petri Net With Data (without variables)");

						context.getProgress().inc();
					}

					// Social perspective (roles)
//				if (configuration.getActivities().contains(ACTIVITY.ROLE_DISCOVERY)) {
//						context.log("Mining roles...");
//						roles = context.tryToFindOrConstructFirstNamedObject(RoleSet.class,
//								PluginNames.SWIMLANE_DISCOVERY, null, null, bpmnModel, log);
//						BPMNDiagram bpmnModelWithSwimlanes = context.tryToFindOrConstructFirstNamedObject(
//								BPMNDiagram.class, PluginNames.DECORATE_BPMN_WITH_SWIMLANES, null, null, bpmnModel,
//								roles);
//						if (bpmnModelWithSwimlanes != null) {
//							bpmnModel = bpmnModelWithSwimlanes;
//						}
//						context.getProgress().inc();
//				}

					// Subprocesses reconstruction
					bpmnModel = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class,
							PluginNames.RECONSTRUCT_SUBPROCESSES, null, null, bpmnModel);
						
					// For social perspective - mine roles inside subprocesses
					if (configuration.getActivities().contains(ACTIVITY.ROLE_DISCOVERY)) {
						context.log("Mining Roles inside Subprocesses...");
						BPMNDiagram bpmnModelWithHierarchicalSwimlanes = context.tryToFindOrConstructFirstNamedObject(
								BPMNDiagram.class, PluginNames.HIERARCHICAL_ROLES_MINER, null, null, bpmnModel, log);
						if (bpmnModelWithHierarchicalSwimlanes != null) {
							bpmnModel = bpmnModelWithHierarchicalSwimlanes;
						}
					}
					// Replace START/END events by START/END activities
					BPMNUtils.replaceEventsByActivities(bpmnModel);
					// Social perspective (social network analysis)
					/*
					 * if
					 * (configuration.getActivities().contains(ACTIVITY.ADVANCED_SNA
					 * )) { // social network analysis
					 * context.log("Mining Advanced Social Network...");
					 * 
					 * try { adSNA = Orchestrator.mineLight(context, log); }
					 * catch (IOException e) { e.printStackTrace(); }
					 * 
					 * context.getProgress().inc(); }
					 */
					
					//---ENHANCE BPMN MODEL WITH ALIGNMENT
					/*Will be available when the dependence on BPMNVisualizer will be resolved*/
					//bpmnModel = context.tryToFindOrConstructFirstNamedObject(BPMNDiagram.class,
					//		PluginNames.MODEL_ENHANCER, null, null, bpmnModel, petriNet, replayResult);
					
					//-------------------------------------
					
				}
			} catch (Exception e) {
				e.printStackTrace();	
				//return null;
			}
		}
		
		System.out.println("Mining time required: " + miningTime);
		System.out.println("Merge time required: " + mergeTime);
		System.out.println("Mining space required: " + miningSpace + " [NOT RELIABLE]");
		System.out.println("Merge space requried: " + mergeSpace + " [NOT RELIABLE]");
		
		context.getGlobalContext().getResourceManager().getResourceForInstance(bpmnModel).setFavorite(true);
		
		MultiperspectiveModel model = new MultiperspectiveModel();
		model.setDiagram(bpmnModel);
		model.setRoleSet(roles);
		model.setDataPetriNet(dataPetriNet);
		model.setDataPetriNetWithoutVariables(dataPetriNetNoVars);
		//model.setAdvancedSocialNetwork(adSNA);
		return model;
	}
	
	private static long usedSpace() {
		return Runtime.getRuntime().totalMemory() - Runtime.getRuntime().freeMemory();
	}
}
