package prompt.multiperspective.plugins.io;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipInputStream;

import org.processmining.contexts.uitopia.annotations.UIImportPlugin;
import org.processmining.framework.abstractplugins.AbstractImportPlugin;
import org.processmining.framework.plugin.PluginContext;
import org.processmining.framework.plugin.annotations.Plugin;
import org.processmining.models.graphbased.directed.bpmn.BPMNDiagram;
import org.processmining.models.graphbased.directed.petrinetwithdata.newImpl.PetriNetWithData;
import org.processmining.plugins.bpmn.Bpmn;
import org.processmining.plugins.bpmn.plugins.BpmnImportPlugin;
import org.processmining.plugins.bpmn.plugins.BpmnSelectDiagramPlugin;

import prompt.multiperspective.plugins.MultiperspectiveModel;
import prompt.roles.data.RoleSet;

/**
 * This class is responsible of importing multiperspecive models.
 * 
 * @author Andrea Burattin
 */
@Plugin(
	name = "Multiperspective Model (.mpm)",
	returnLabels = { "Imported Multiperspective Model" },
	returnTypes = { MultiperspectiveModel.class },
	parameterLabels = { "File" },
	userAccessible = true
)
@UIImportPlugin(
	description = "Multiperspective Model (.mpm)",
	extensions = { "mpm" }
)
public class MultiperspectiveModelImport extends AbstractImportPlugin {
	
	/**
	 * The actual import method.
	 */
	@Override
	protected Object importFromStream(PluginContext context, InputStream input, String filename, long fileSizeInBytes) throws Exception {
		try {
			MultiperspectiveModel model = new MultiperspectiveModel(null, null);
			
			ZipInputStream zin = new ZipInputStream(input);
			ZipEntry entry;
			
			File tempFile = File.createTempFile("dummy", ".file");
			tempFile.deleteOnExit();
			
			while((entry = zin.getNextEntry()) != null) {
				int n; byte[] buf = new byte[1024];
				FileOutputStream fileoutputstream = new FileOutputStream(tempFile);
				while ((n = zin.read(buf, 0, 1024)) > -1) {
					fileoutputstream.write(buf, 0, n);
				}
				fileoutputstream.close();
				zin.closeEntry();
				
				if (entry.getName().equals(MultiperspectiveModelIOUtils.ROLESET_FILENAME)) {
					System.out.println("Loading roleset file...");
					RoleSet rs = RoleSet.loadFromFile(tempFile);
					model.setRoleSet(rs);

				} else if (entry.getName().equals(MultiperspectiveModelIOUtils.BPMN_FILENAME)) {
					System.out.println("Loading BPMN model...");
					BpmnImportPlugin importer = new BpmnImportPlugin();
					BpmnSelectDiagramPlugin selector = new BpmnSelectDiagramPlugin();
					
					Bpmn bpmn = (Bpmn) importer.importFile(context, tempFile);
					BPMNDiagram diagram = selector.selectDefault(context, bpmn);
					model.setDiagram(diagram);

				} else if (entry.getName().equals(MultiperspectiveModelIOUtils.DATAPETRINET_FILENAME)) {
					System.out.println("Loading data Petri net...");
					PetriNetWithData net = context.tryToFindOrConstructFirstObject(PetriNetWithData.class, null, null, tempFile);
					model.setDataPetriNet(net);

				} else if (entry.getName().equals(MultiperspectiveModelIOUtils.DATAPETRINET_NOVARS_FILENAME)) {
					System.out.println("Loading data Petri net without variables...");
					PetriNetWithData net = context.tryToFindOrConstructFirstObject(PetriNetWithData.class, null, null, tempFile);
					model.setDataPetriNetWithoutVariables(net);
					
				} /*else if (entry.getName().equals(MultiperspectiveModelIOUtils.ADVANCED_SNA_FILENAME) && tempFile.length() != 0) {
					System.out.println("Loading advanced SNA...");
					AdvancedSocialNetwork adSNA = new AdvancedSocialNetwork(tempFile);
					model.setAdvancedSocialNetwork(adSNA);
				}*/
			}
			
			context.getFutureResult(0).setLabel("Multiperspective Model imported from " + filename);
			return model;
			
		} catch (IOException e) {
			e.printStackTrace();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return null;
	}
}
